from typing import Tuple, Callable, TypeVar

from amino import do, Do, Either, Maybe, _, curried, L, List, Nothing, Just
from amino.state import EitherState, State
from amino.case import Case
from amino.logging import module_log
from amino.lenses.lens import lens
from amino.boolean import true
from chiasma.data.view_tree import (map_panes, ViewTree, find_in_view_tree, LayoutNode, map_layouts, map_layout_nodes,
                                    map_pane_nodes)
from chiasma.util.id import IdentSpec
from chiasma.open_pane import pane_open_layout_hook, mod_pane_open, pane_node_open
from chiasma.mod_pane import mod_pane

from myo.util import Ident
from myo.ui.data.ui_data import UiData
from myo.ui.data.window import Window
from myo.ui.data.space import Space
from myo.ui.data.view import Layout, Pane, has_ident
from myo.ui.data.view_tree import MyoViewTree, MyoLayoutNode, MyoPaneNode

A = TypeVar('A')
log = module_log()


def find_in_window(
        pred: Callable[[MyoViewTree], Maybe[A]],
        window: Window,
) -> Maybe[Tuple[Window, MyoViewTree]]:
    return find_in_view_tree(pane=pred, layout=pred)(window.layout).map(lambda pane: (window, pane))


def find_in_windows(
        pred: Callable[[MyoViewTree], Maybe[A]],
        space: Space,
) -> Maybe[Tuple[Space, Window, MyoViewTree]]:
    a = space.windows.find_map(L(find_in_window)(pred, _))
    return a.map2(lambda w, p: (space, w, p))


def find_in_spaces(pred: Callable[[MyoViewTree], Maybe[A]], spaces: List[Space]) -> None:
    return spaces.find_map(lambda a: find_in_windows(pred, a))


@do(State[UiData, Maybe[Tuple[Space, Window, MyoViewTree]]])
def find_in_ui(pred: Callable[[MyoViewTree], Maybe[A]]) -> Do:
    spaces = yield State.inspect(_.spaces)
    return find_in_spaces(pred, spaces)


spaces_layout_lens = lens.Each().windows.Each().layout
uidata_layout_lens = lens.spaces & spaces_layout_lens


def map_window_trees(mod: Callable[[MyoViewTree], MyoViewTree]) -> EitherState[str, UiData, None]:
    return EitherState.modify(uidata_layout_lens.modify(mod))


def ui_modify_layout_nodes(
        pred: Callable[[MyoLayoutNode], bool],
        mod: Callable[[MyoLayoutNode], MyoLayoutNode],
) -> EitherState[str, UiData, None]:
    return map_window_trees(map_layout_nodes(pred, mod))


def ui_modify_pane_nodes(
        pred: Callable[[Pane], bool],
        mod: Callable[[MyoPaneNode], MyoPaneNode],
) -> EitherState[str, UiData, None]:
    return map_window_trees(map_pane_nodes(pred, mod))


def ui_modify_layout_node(ident: Ident, mod: Callable[[MyoLayoutNode], MyoLayoutNode]) -> EitherState[str, UiData, Window]:
    return ui_modify_layout_nodes(lambda a: has_ident(ident)(a.data), mod)


def ui_modify_pane_node(ident: Ident, mod: Callable[[MyoPaneNode], MyoPaneNode]) -> EitherState[str, UiData, Window]:
    return ui_modify_pane_nodes(lambda a: has_ident(ident)(a.data), mod)


def ui_modify_layouts(pred: Callable[[Layout], bool], mod: Callable[[Layout], Layout]) -> EitherState[str, UiData, None]:
    return map_window_trees(map_layouts(pred, mod))


def ui_modify_panes(pred: Callable[[Pane], bool], mod: Callable[[Pane], Pane]) -> EitherState[str, UiData, None]:
    return map_window_trees(map_panes(pred, mod))


def ui_modify_layout(ident: Ident, mod: Callable[[Layout], Layout]) -> EitherState[str, UiData, Window]:
    return ui_modify_layouts(has_ident(ident), mod)


def ui_modify_pane(ident: Ident, mod: Callable[[Pane], Pane]) -> EitherState[str, UiData, Window]:
    return ui_modify_panes(has_ident(ident), mod)


def map_panes_in_spaces(
        pred: Callable[[Pane], bool],
        update: Callable[[Pane], Pane],
) -> List[Space]:
    return spaces_layout_lens.modify(map_panes(pred, update))


class insert_pane(Case, alg=ViewTree):

    def __init__(self, pane: Pane, layout: Ident) -> None:
        self.pane = pane
        self.layout = layout

    def layout_node(self, node: LayoutNode) -> Maybe[ViewTree]:
        return (
            Just(node.append1.sub(ViewTree.pane(self.pane)))
            if node.data.ident == self.layout else
            node.sub.find_map(self) / node.replace_sub
        )

    def case_default(self, node: ViewTree) -> Maybe[ViewTree]:
        return Nothing


@curried
def insert_pane_into_window(pane: Pane, layout: Ident, window: Window) -> Maybe[Window]:
    return insert_pane(pane, layout)(window.layout) / window.set.layout


@curried
def insert_pane_into_space(pane: Pane, layout: Ident, space: Space) -> Maybe[Space]:
    return space.windows.find_map(insert_pane_into_window(pane, layout)) / space.replace_window


@curried
def insert_pane_into_ui(pane: Pane, layout: Ident, ui: UiData) -> Either[str, UiData]:
    return (
        ui.spaces.find_map(insert_pane_into_space(pane, layout))
        .map(ui.replace_space)
        .to_either(lambda: f'no layout with ident {layout}')
    )


def open_or_toggle_minimized(node: MyoPaneNode) -> MyoPaneNode:
    l = (
        lens.data.state.minimized.modify(lambda a: ~a)
        if node.data.open else
        pane_node_open.set(true)
    )
    return l(node)


def open_or_toggle_pane(spec: IdentSpec) -> Callable[[MyoViewTree], Either[str, MyoViewTree]]:
    return mod_pane(mod_pane_open(spec, open_or_toggle_minimized), pane_open_layout_hook)


__all__ = ('find_in_window', 'find_in_windows', 'find_in_ui', 'ui_modify_pane', 'insert_pane_into_ui',
           'ui_modify_layout',)



from amino import List, Map

from ribosome.config.config import Config

from myo.env import Env
from myo.components.core.config import core
from myo.components.command.config import command
from myo.components.tmux.config import tmux
from myo.components.ui.config import ui
from myo.components.core.compute.init import init

myo_config: Config = Config.cons(
    name='myo',
    prefix='myo',
    state_ctor=Env.cons,
    components=Map(core=core, command=command, ui=ui, tmux=tmux),
    core_components=List('core'),
    default_components=List('command', 'ui', 'tmux'),
    init=init,
)

__all__ = ('myo_config')



from amino import List, Map, Just
from amino.boolean import true

from ribosome.config.component import Component
from ribosome.rpc.api import rpc
from ribosome.util.doc.data import DocBlock

from myo.config.component import MyoComponent
from myo.ui.data.ui_data import UiData
from myo.components.ui.compute.open_pane import open_pane
from myo.components.ui.compute.close_pane import close_pane
from myo.components.ui.compute.minimize_pane import minimize_pane
from myo.components.ui.compute.create_pane import create_pane
from myo.components.ui.compute.init import init
from myo.components.ui.compute.info import ui_info
from myo.components.ui.compute.toggle_pane import toggle_pane
from myo.components.ui.compute.toggle_layout import toggle_layout
from myo.components.ui.compute.kill_pane import kill_pane
from myo.components.ui.compute.focus import focus

create_pane_help = DocBlock.string('''Configure a new pane that can be spawned with `MyoOpenPane`.''')
create_pane_json_help = Map(
    layout='containing layout name',
    ident='name to be used for commands like MyoOpenPane',
    min_size='minimum size integer',
    max_size='maximum size integer',
    fixed_size='fixed size, combination of min and max',
    minimized_size='size after calling MyoMinimizePane',
    weight='amount of growth this pane gets when there is space left above the minimums',
    position='absolute position in the layout',
    minimized='whether to create the pane in minimized state',
)


ui = Component.cons(
    'ui',
    state_type=UiData,
    rpc=List(
        rpc.write(create_pane).conf(json=true, help=create_pane_help, json_help=Just(create_pane_json_help)),
        rpc.write(open_pane).conf(json=true),
        rpc.write(close_pane),
        rpc.write(toggle_pane),
        rpc.write(minimize_pane),
        rpc.write(kill_pane),
        rpc.write(toggle_layout),
        rpc.read(ui_info),
        rpc.read(focus),
    ),
    config=MyoComponent.cons(info=ui_info, init=init),
)

__all__ = ('ui',)



from amino import List

from ribosome.config.component import Component
from ribosome.rpc.api import rpc

from myo.components.core.compute.init import init
from myo.components.core.data import CoreData
from myo.components.core.compute.info import info
from myo.components.core.compute.vim_leave import vim_leave


core = Component.cons(
    'core',
    state_type=CoreData,
    rpc=List(
        rpc.write(init),
        rpc.write(info),
        rpc.autocmd(vim_leave, sync=True),
    ),
)

__all__ = ('core',)



from amino import Path, do, Do, IO

from ribosome.nvim.io.state import NS

from myo.components.command.compute.tpe import CommandRibosome


@do(NS[CommandRibosome, str])
def project_relative_path(path: Path) -> Do:
    rel = yield NS.from_io(IO.delay(path.relative_to, Path.cwd()).recover(lambda a: path))
    return str(rel)


__all__ = ('project_relative_path',)



from typing import Callable

from amino import Maybe, Do

from ribosome.compute.api import prog
from ribosome.compute.program import Program
from ribosome.compute.prog import Prog

from chiasma.util.id import Ident

from myo.ui.ui import Ui
from myo.components.ui.compute.pane import view_owners


@prog.do(Program)
def ui_handler(ident: Ident, handler: Callable[[Ui], Maybe[Program]], desc: str) -> Do:
    space, window, owner = yield view_owners(ident)
    yield Prog.m(handler(owner), lambda: f'ui `{type(owner)}` has no handler for `{desc}`')


__all__ = ('ui_handler',)



from typing import TypeVar

from chiasma.util.id import IdentSpec, Ident, ensure_ident

from ribosome.nvim.io.state import NS
from ribosome.compute.prog import Prog

D = TypeVar('D')


def ensure_ident_ns(spec: IdentSpec) -> NS[D, Ident]:
    return NS.e(ensure_ident(spec))


def ensure_ident_prog(spec: IdentSpec) -> Prog[Ident]:
    return Prog.e(ensure_ident(spec))


__all__ = ('ensure_ident_ns', 'ensure_ident_prog',)



from amino import _, List, Nil, Either, Maybe
from amino.dat import Dat

from myo.util import Ident
from myo.data.command import Command, TestLineParams


class Env(Dat['Env']):

    @staticmethod
    def cons(
            test_params: TestLineParams=None,
    ) -> 'Env':
        return Env(Maybe.optional(test_params))

    def __init__(
            self,
            test_params: Maybe[TestLineParams],
    ) -> None:
        self.test_params = test_params

__all__ = ('Env',)



from amino import List, Nil, do, Do
from amino.case import Case

from myo.components.command.compute.tpe import CommandRibosome
from myo.output.parser.python import PythonLine, PythonEvent, StackFrameEvent, ErrorEvent, FileLine
from myo.output.data.output import OutputEvent, Location, OutputLine
from myo.components.command.compute.parse_handlers import PathFormatter
from myo.output.data.report import DisplayLine, PlainDisplayLine

from ribosome.nvim.io.state import NS


@do(NS[CommandRibosome, List[DisplayLine]])
def format_location(file: OutputLine[FileLine], location: Location, path_formatter: PathFormatter) -> Do:
    fun = file.meta.fun.map(lambda a: f' {a}').get_or_strict('')
    path = yield path_formatter(location.path)
    return List(PlainDisplayLine(f'{path}  {location.line + 1}{fun}'))


class python_report_event(Case[PythonEvent, NS[CommandRibosome, List[DisplayLine]]], alg=PythonEvent):

    def __init__(self, path_formatter: PathFormatter, event: OutputEvent[PythonLine, PythonEvent]) -> None:
        self.path_formatter = path_formatter
        self.event = event

    @do(NS[CommandRibosome, List[DisplayLine]])
    def stack_frame(self, event: StackFrameEvent) -> Do:
        loc_line = yield (
            self.event.location
            .map(lambda a: format_location(event.file, a, self.path_formatter))
            .get_or_strict(NS.pure(Nil))
        )
        code_line = event.code.map(lambda a: PlainDisplayLine(a.text)).to_list
        return loc_line + code_line

    def error(self, event: ErrorEvent) -> NS[CommandRibosome, List[DisplayLine]]:
        return NS.pure(List(PlainDisplayLine(event.line.text)))


def python_report(
        path_formatter: PathFormatter,
        output: OutputEvent[PythonLine, PythonEvent]
) -> NS[CommandRibosome, List[DisplayLine]]:
    return python_report_event(path_formatter, output)(output.meta)


__all__ = ('python_report',)



from setuptools import setup, find_packages

version_parts = (1, 0, 0, 'a', 38)
version = '.'.join(map(str, version_parts))

setup(  # type: ignore
    name='myo',
    description='command and layout management for neovim',
    version=version,
    author='Torsten Schmits',
    author_email='torstenschmits@gmail.com',
    license='MIT',
    url='https://github.com/tek/myo',
    packages=find_packages(exclude=['unit', 'unit.*', 'integration', 'integration.*', 'test', 'test.*']),
    install_requires=[
        'ribosome==13.0.1a6',
        'chiasma~=0.1.0.a28',
        'psutil==5.3.1',
        'networkx==2.0',
        'lark-parser==0.6.5',
    ],
    tests_require=[
        'kallikrein~=0.22.a17',
    ],
)



from typing import TypeVar, Callable

from ribosome.nvim.io.state import NS
from ribosome.compute.ribosome_api import Ribo

from chiasma.util.id import Ident

from amino import do, Do, List, __, Maybe, IO, Lists, Just, Nothing, Nil
from amino.logging import module_log
from amino.case import Case

from myo.output.data.output import OutputEvent
from myo.components.command.data import CommandData
from myo.data.command import Command, Interpreter, ShellInterpreter, CommandConfig
from myo.components.command.compute.tpe import CommandRibosome
from myo.command.history import most_recent_command
from myo.output.main import parse_with_langs
from myo.output.config import LangConfig, ParseConfig
from myo.output.configs import default_lang_configs, global_config_defaults
from myo.components.command.compute.parsed_output import ParsedOutput
from myo.components.command.compute.parse_handlers import ParseHandlers
from myo.config.settings import builtin_output_config

log = module_log()
A = TypeVar('A')
D = TypeVar('D')


def command_config_option(cmd: Maybe[CommandConfig], shell: Maybe[CommandConfig]
                          ) -> Callable[[Callable[[CommandConfig], A], Callable[[A, A], A]], Maybe[A]]:
    def command_config_option(get: Callable[[CommandConfig], A], combine: Callable[[A, A], A]) -> Maybe[A]:
        primary = cmd.flat_map(get)
        secondary = shell.flat_map(get)
        return (
            primary
            .map(
                lambda a:
                secondary.map(lambda b: combine(a, b))
                .get_or_strict(a)
            )
            .o(secondary)
        )
    return command_config_option


def command_conf_by_ident(ident: Ident) -> NS[CommandRibosome, Maybe[CommandConfig]]:
    return Ribo.inspect_comp(lambda a: a.command_configs.find(lambda b: b.ident == ident))


def find_lang_config(configs: List[LangConfig], name: str) -> Maybe[ParseConfig]:
    return configs.find(lambda a: a.name == name).map(lambda a: a.parse)


def lang_config(use_builtins: bool) -> Callable[[str], NS[CommandRibosome, Maybe[ParseConfig]]]:
    @do(NS[CommandRibosome, Maybe[ParseConfig]])
    def lang_config(lang: str) -> Do:
        configs = yield Ribo.inspect_comp(lambda a: a.lang_configs)
        specific = find_lang_config(configs, lang)
        return (
            specific.or_else_call(find_lang_config, default_lang_configs, lang)
            if use_builtins else
            specific
        )
    return lang_config


@do(NS[CommandRibosome, ParseConfig])
def parse_config(cmd: Command, shell: Maybe[Command]) -> Do:
    use_builtins = yield Ribo.setting(builtin_output_config)
    cmd_conf = yield command_conf_by_ident(cmd.ident)
    shell_conf = yield shell.map(lambda a: command_conf_by_ident(a.ident)).get_or(NS.pure, Nothing)
    shell_langs = shell.map(lambda a: a.langs).get_or_strict(Nil)
    langs = (cmd.langs + shell_langs).distinct
    lang_confs = yield langs.flat_traverse((lang_config)(use_builtins), NS)
    cmd_lang_conf = ParseConfig.cons(langs=langs)
    global_conf = global_config_defaults.parse
    local_confs = (cmd_conf.to_list + shell_conf.to_list).map(lambda a: a.parse)
    nonglobal_confs = (local_confs + lang_confs).cons(cmd_lang_conf)
    confs = (
        nonglobal_confs.cat(global_conf)
        if use_builtins else
        nonglobal_confs
    )
    yield NS.pure(confs.fold(ParseConfig))


@do(NS[CommandRibosome, List[OutputEvent]])
def parse_with_config(output: List[str], langs: List[str]) -> Do:
    yield NS.e(parse_with_langs(output, langs))


@do(NS[CommandRibosome, ParsedOutput])
def parse_output(cmd: Command, output: List[str], shell: Maybe[Command]) -> Do:
    config = yield parse_config(cmd, shell)
    handlers = yield NS.e(ParseHandlers.from_config(config))
    log.debug(f'parsing output with {handlers}')
    events = yield parse_with_config(output, config.langs)
    filtered = yield handlers.filter.fold_m(NS.pure(events))(lambda z, f: f(z))
    return ParsedOutput(handlers, events, filtered)


@do(NS[CommandData, List[str]])
def cmd_output(ident: Ident) -> Do:
    cmd_log = yield NS.inspect_either(__.log_by_ident(ident))
    text = yield NS.from_io(IO.delay(cmd_log.read_text))
    return Lists.lines(text)


class shell_for_command(Case[Interpreter, NS[CommandData, Maybe[Command]]], alg=Interpreter):

    @do(NS[CommandData, Maybe[Command]])
    def shell_interpreter(self, interpreter: ShellInterpreter) -> Do:
        shell = yield NS.inspect_either(lambda a: a.command_by_ident(interpreter.target))
        return Just(shell)

    def case_default(self, interpreter: Interpreter) -> NS[CommandData, Maybe[Command]]:
        return NS.pure(Nothing)


@do(NS[CommandRibosome, ParsedOutput])
def parse_command(cmd: Command, shell: Maybe[Command]) -> Do:
    output = yield Ribo.zoom_comp(cmd_output(cmd.ident))
    yield parse_output(cmd, output, shell)


@do(NS[CommandRibosome, ParsedOutput])
def parse_most_recent() -> Do:
    cmd = yield Ribo.zoom_comp(most_recent_command())
    shell = yield Ribo.zoom_comp(shell_for_command.match(cmd.interpreter))
    yield parse_command(cmd, shell)


__all__ = ('parse_command', 'parse_most_recent',)



from typing import Callable, TypeVar, Generic

from amino import Dat, List, Maybe, Either, Path, Nothing, Nil, do, Do

from ribosome.nvim.io.state import NS
from ribosome.nvim.syntax.syntax import Syntax

from myo.components.command.compute.tpe import CommandRibosome
from myo.output.data.output import OutputEvent
from myo.output.config import ParseConfig
from myo.output.data.report import DisplayLine, PlainDisplayLine

A = TypeVar('A')
B = TypeVar('B')
OutputFilter = Callable[[List[OutputEvent[A, B]]], List[OutputEvent[A, B]]]
FirstError = Callable[[List[OutputEvent[A, B]]], NS[CommandRibosome, int]]
PathFormatter = Callable[[Path], NS[CommandRibosome, str]]
Reporter = Callable[[PathFormatter, OutputEvent[A, B]], NS[CommandRibosome, List[List[DisplayLine]]]]
SyntaxCons = Callable[[], NS[CommandRibosome, Syntax]]


def default_first_error(output: List[OutputEvent[A, B]]) -> NS[CommandRibosome, int]:
    return NS.pure(0)


def default_path_formatter(path: Path) -> NS[CommandRibosome, str]:
    return NS.pure(str(path))


def verbatim_report(formatter: PathFormatter, event: OutputEvent[A, B]) -> NS[CommandRibosome, List[str]]:
    return NS.pure(event.lines.map(lambda a: PlainDisplayLine(a.text)))


def no_syntax() -> NS[CommandRibosome, Syntax]:
    return NS.pure(Syntax.cons())


class ParseHandlers(Generic[A, B], Dat['ParseHandlers[A, B]']):

    @staticmethod
    def from_config(config: ParseConfig) -> Either[str, 'ParseHandlers']:
        return ParseHandlers.resolve(
            config.filter,
            config.first_error,
            config.path_formatter,
            config.reporter,
            config.syntax,
        )

    @staticmethod
    @do(Either[str, 'ParseHandlers'])
    def resolve(
            filter: List[str],
            first_error: Maybe[str],
            path_formatter: Maybe[str],
            reporter: Maybe[str],
            syntax: Maybe[str],
    ) -> Do:
        filter_handlers = yield filter.traverse(Either.import_path, Either)
        first_error_handler = yield first_error.traverse(Either.import_path, Either)
        path_formatter_handler = yield path_formatter.traverse(Either.import_path, Either)
        reporter_handler = yield reporter.traverse(Either.import_path, Either)
        syntax_handler = yield syntax.traverse(Either.import_path, Either)
        return ParseHandlers.cons(
            filter_handlers,
            first_error_handler,
            path_formatter_handler,
            reporter_handler,
            syntax_handler,
        )

    @staticmethod
    def cons(
            filter: List[OutputFilter]=Nil,
            first_error: Maybe[FirstError]=Nothing,
            path_formatter: Maybe[PathFormatter]=Nothing,
            reporter: Maybe[Reporter]=Nothing,
            syntax: Maybe[SyntaxCons]=Nothing
    ) -> 'ParseHandlers':
        return ParseHandlers(
            filter,
            first_error.get_or_strict(default_first_error),
            path_formatter.get_or_strict(default_path_formatter),
            reporter.get_or_strict(verbatim_report),
            syntax.get_or_strict(no_syntax),
        )


    def __init__(
            self,
            filter: List[OutputFilter],
            first_error: FirstError,
            path_formatter: PathFormatter,
            reporter: Reporter,
            syntax: SyntaxCons,
    ) -> None:
        self.filter = filter
        self.first_error = first_error
        self.path_formatter = path_formatter
        self.reporter = reporter
        self.syntax = syntax


__all__ = ('ParseHandlers',)



from typing import Union

from networkx import DiGraph

from amino import List, Path, do, Either, Try, Do, Maybe, Regex, Nil, Nothing, ADT, Just, Dat
from amino.util.numeric import parse_int
from amino.logging import module_log

from myo.output.data.output import OutputLine, OutputEvent, Location
from myo.output.parser.base import EdgeData, Parser

log = module_log()


class ScalaLine(ADT['ScalaLine']):
    pass


class FileLine(ScalaLine):

    @staticmethod
    @do(Either[str, 'FileLine'])
    def cons(
            path: Union[str, Path],
            line: Union[str, int],
            col: Union[str, int]=None,
            error: str='',
            tag: str='',
    ) -> Do:
        path_p = yield Try(Path, path)
        line_i = yield parse_int(line)
        col_i = yield parse_int(0 if col is None else col)
        return FileLine(path_p, line_i - 1, col_i, error, tag)

    def __init__(self, path: Path, line: int, col: int, error: str, tag: str) -> None:
        self.path = path
        self.line = line
        self.col = col
        self.error = error
        self.tag = tag


class InfoLine(ScalaLine):

    @staticmethod
    def cons(message: str, ws: str='', tag: str='') -> 'InfoLine':
        return InfoLine(message, len(ws), tag)

    def __init__(self, message: str, indent: int, tag: str) -> None:
        self.message = message
        self.indent = indent
        self.tag = tag


class CodeLine(ScalaLine):

    def __init__(self, code: str, indent: int, tag: str) -> None:
        self.code = code
        self.indent = indent
        self.tag = tag


class ColLine(ScalaLine):

    @staticmethod
    def cons(ws: str, tag: str='') -> 'ColLine':
        return ColLine(len(ws), tag)

    def __init__(self, col: int, tag: str) -> None:
        self.col = col
        self.tag = tag


_msg_type = '\[(?P<tag>\w+)\] '
file_edge: EdgeData[FileLine] = EdgeData(
    regex=Regex(f'^{_msg_type}\s*(?P<path>[^:]+):(?P<line>\d+):((?P<col>\d+):)?\s*(?P<error>.*?);?$'),
    cons_output_line=FileLine.cons,
)
info_edge = EdgeData.strict(
    regex=Regex(f'^{_msg_type}(?P<ws>\s*)(?P<message>\s*.+)'),
    cons_output_line=InfoLine.cons,
)
col_edge = EdgeData.strict(
    regex=Regex(f'^{_msg_type}(?P<ws>\s*)\^\s*'),
    cons_output_line=ColLine.cons,
)


class ScalaEvent(Dat['ScalaEvent']):

    @staticmethod
    def cons(
            file: OutputLine[FileLine],
            info: List[OutputLine[InfoLine]]=Nil,
            code: Maybe[OutputLine[CodeLine]]=Nothing,
            col: Maybe[OutputLine[ColLine]]=Nothing,
    ) -> 'ScalaEvent':
        return ScalaEvent(file, info, code, col)

    def __init__(
            self,
            file: OutputLine[FileLine],
            info: List[OutputLine[InfoLine]],
            code: Maybe[OutputLine[CodeLine]],
            col: Maybe[OutputLine[ColLine]],
    ) -> None:
        self.file = file
        self.info = info
        self.code = code
        self.col = col


def scala_graph() -> DiGraph:
    g = DiGraph()
    g.add_edge('start', 'file', data=file_edge)
    g.add_edge('file', 'info', data=info_edge)
    g.add_edge('info', 'info', data=info_edge)
    g.add_edge('info', 'col', data=col_edge, weight=1)
    return g


@do(Maybe[OutputEvent[ScalaLine, ScalaEvent]])
def scala_event(lines: List[OutputLine[ScalaLine]]) -> Do:
    col = lines.find(lambda a: isinstance(a.meta, ColLine))
    file = yield lines.find(lambda a: isinstance(a.meta, FileLine))
    info = lines.filter(lambda a: isinstance(a.meta, InfoLine))
    code, infos = (
        info.detach_last
        .map2(lambda c, e: (Just(OutputLine(c.text, CodeLine(c.meta.message, c.meta.indent, c.meta.tag), c.indent)), e))
        .get_or_strict((Nothing, Nil))
    )
    location = Location(file.meta.path, file.meta.line, col.map(lambda a: a.meta.col).get_or_strict(0))
    return OutputEvent.cons(ScalaEvent(file, infos, code, col), lines, Just(location))


def scala_events(lines: List[OutputLine[ScalaLine]]) -> List[OutputEvent[ScalaLine, ScalaEvent]]:
    return scala_event(lines).to_list


scala_parser = Parser(scala_graph(), scala_events)

__all__ = ('scala_parser',)



from amino import List, Just, Nothing, Maybe

from ribosome.config.component import Component
from ribosome.compute.program import Program
from ribosome.rpc.api import rpc

from myo.components.tmux.data import TmuxData

from myo.components.tmux.compute.run import tmux_run_command, tmux_can_run
from myo.config.component import MyoComponent
from myo.ui.ui import Ui
from myo.components.tmux.compute.owns_pane import tmux_owns_view
from myo.components.tmux.compute.render import tmux_render
from myo.command.run_task import RunTask
from myo.components.tmux.compute.create_vim_pane import create_vim_pane
from myo.components.tmux.compute.info import tmux_info
from myo.components.tmux.compute.quit import tmux_quit
from myo.components.tmux.compute.kill_pane import tmux_kill_pane
from myo.components.tmux.compute.pane_open import tmux_window_pane_open
from myo.components.tmux.compute.focus import tmux_focus_pane


def run_handler_for(task: RunTask) -> Maybe[Program]:
    return Just(tmux_run_command) if tmux_can_run(task) else Nothing


def create_vim_pane_handler() -> Maybe[Program]:
    return Just(create_vim_pane)


tmux_ui = Ui.cons(tmux_owns_view, tmux_render, kill_pane=tmux_kill_pane, window_pane_open=tmux_window_pane_open,
                  focus_pane=tmux_focus_pane)
tmux: Component[TmuxData, MyoComponent] = Component.cons(
    'tmux',
    state_type=TmuxData,
    rpc=List(
        rpc.write(tmux_render),
        rpc.write(tmux_info),
        rpc.write(tmux_quit),
    ),
    config=MyoComponent.cons(
        run_handler_for,
        create_vim_pane_handler,
        info=tmux_info,
        ui=tmux_ui,
        quit=tmux_quit,
    ),
)

__all__ = ('tmux',)



from typing import Generic, TypeVar

from amino import Dat, List

from myo.components.command.compute.parse_handlers import ParseHandlers
from myo.output.data.output import OutputEvent

A = TypeVar('A')
B = TypeVar('B')


class ParsedOutput(Generic[A, B], Dat['ParsedOutput[A, B]']):

    def __init__(
            self,
            handlers: ParseHandlers,
            original: List[OutputEvent[A, B]],
            filtered: List[OutputEvent[A, B]],
    ) -> None:
        self.handlers = handlers
        self.original = original
        self.filtered = filtered


__all__ = ('ParsedOutput',)



from amino import do, Do
from amino.logging import module_log

from ribosome.nvim.io.state import NS
from ribosome.compute.api import prog
from ribosome.compute.ribosome_api import Ribo

from myo.components.command.compute.tpe import CommandRibosome
from myo.components.command.compute.history import restore_history
from myo.config.settings import load_history

log = module_log()


@prog
@do(NS[CommandRibosome, None])
def command_init() -> Do:
    want_load_history = yield Ribo.setting(load_history)
    if want_load_history:
        yield Ribo.zoom_comp(restore_history())


__all__ = ('command_init',)



from typing import TypeVar

from amino import Dat, List, ADT, do, Do, Maybe
from amino.state import State
from amino.case import Case

from myo.output.data.output import OutputEvent, OutputLine, Location

A = TypeVar('A')
B = TypeVar('B')


class DisplayLine(ADT['DisplayLine']):
    pass


class PlainDisplayLine(DisplayLine):

    def __init__(self, text: str) -> None:
        self.text = text


class SyntaxDisplayLine(DisplayLine):

    def __init__(self, text: str, syntax: str) -> None:
        self.text = text
        self.syntax = syntax


class ReportLine(ADT['ReportLine']):
    pass


class PlainReportLine(ReportLine):

    def __init__(self, text: str) -> None:
        self.text = text


class EventReportLine(ReportLine):

    @staticmethod
    def cons(
            line: DisplayLine,
            event: int,
            location: Location=None,
    ) -> 'EventReportLine':
        return EventReportLine(
            line,
            event,
            Maybe.optional(location),
        )

    def __init__(self, line: DisplayLine, event: int, location: Maybe[Location]) -> None:
        self.line = line
        self.event = event
        self.location = location


class ParseReport(Dat['ParseReport']):

    @staticmethod
    def cons(
            lines: List[ReportLine],
            event_indexes: List[int],
    ) -> 'ParseReport':
        return ParseReport(
            lines,
            event_indexes,
        )

    def __init__(self, lines: List[ReportLine], event_indexes: List[int]) -> None:
        self.lines = lines
        self.event_indexes = event_indexes


def output_line_report(line: OutputLine[A], event: int, location: Maybe[Location]) -> List[ReportLine]:
    return List(EventReportLine(line.text, event, location))


def event_report(index: int, event: OutputEvent[A, B]) -> List[ReportLine]:
    return (
        (event.head / PlainReportLine) +
        (event.lines.flat_map(lambda a: output_line_report(a, index, event.location)))
    )


@do(State[int, int])
def event_index(lines: List[ReportLine]) -> Do:
    count = lines.length
    current = yield State.get()
    yield State.set(current + count)
    return current


class line_report(Case[DisplayLine, str], alg=DisplayLine):

    def plain(self, line: PlainDisplayLine) -> str:
        return line.text

    def syntax(self, line: SyntaxDisplayLine) -> str:
        return line.text


def format_report(report: ParseReport) -> List[str]:
    return report.lines.map(lambda a: line_report.match(a.line))


__all__ = ('ReportLine', 'format_report',)



from amino import Dat, Maybe, Boolean
from amino.boolean import false

from chiasma.util.id import Ident

from ribosome.compute.program import Program
from ribosome.compute.api import prog


@prog.strict
def owns_no_view(ident: Ident) -> Boolean:
    return false


class Ui(Dat['Ui']):

    @staticmethod
    def cons(
            owns_view: Program[bool]=owns_no_view,
            render: Program=None,
            open_pane: Program=None,
            open_window: Program=None,
            open_space: Program=None,
            kill_pane: Program[bool]=None,
            window_pane_open: Program[bool]=None,
            focus_pane: Program[None]=None,
    ) -> 'Ui':
        return Ui(
            owns_view,
            Maybe.optional(render),
            Maybe.optional(open_pane),
            Maybe.optional(open_window),
            Maybe.optional(open_space),
            Maybe.optional(kill_pane),
            Maybe.optional(window_pane_open),
            Maybe.optional(focus_pane),
        )

    def __init__(
            self,
            owns_view: Program[bool],
            render: Maybe[Program],
            open_pane: Maybe[Program],
            open_window: Maybe[Program],
            open_space: Maybe[Program],
            kill_pane: Maybe[Program[bool]],
            window_pane_open: Maybe[Program[bool]],
            focus_pane: Maybe[Program[None]],
    ) -> None:
        self.owns_view = owns_view
        self.render = render
        self.open_pane = open_pane
        self.open_window = open_window
        self.open_space = open_space
        self.kill_pane = kill_pane
        self.window_pane_open = window_pane_open
        self.focus_pane = focus_pane


__all__ = ('Ui',)



from amino import do, Do, Either, Just, env
from amino.logging import module_log
from amino.lenses.lens import lens

from ribosome.compute.api import prog
from ribosome.nvim.io.state import NS
from ribosome.compute.ribosome_api import Ribo

from chiasma.io.state import TS
from myo.components.tmux.data import TmuxData
from chiasma.util.id import Ident
from chiasma.commands.pane import pane, parse_pane_id
from chiasma.data.pane import Pane
from chiasma.data.session import Session
from chiasma.data.window import Window
from chiasma.util.pid import discover_pane_by_pid

from myo.components.tmux.tpe import TmuxRibosome
from myo.config.settings import vim_tmux_pane

log = module_log()


@do(TS[TmuxData, None])
def insert_vim_pane(
        ident: Ident,
        pane_id: Either[str, int],
        vim_pid: int,
) -> Do:
    pane_data = yield TS.lift(pane_id.cata(lambda err: discover_pane_by_pid(vim_pid), pane))
    vim_pane = Pane.cons(ident, id=pane_data.id)
    yield TS.modify(
        lambda s:
        s
        .append1.panes(vim_pane)
        .append1.windows(Window.cons(ident, id=pane_data.window_id))
        .append1.sessions(Session.cons(ident, id=pane_data.session_id))
    ).zoom(lens.views)
    yield TS.modify(lambda s: s.set.vim_pane(Just(vim_pane.ident)))


@prog
@do(NS[TmuxRibosome, None])
def create_vim_pane(ident: Ident, vim_pid: int, use_env_pane: bool) -> Do:
    env_pane = env.get('TMUX_PANE').flat_map(parse_pane_id)
    override = yield Ribo.setting_e(vim_tmux_pane)
    strict = override.o(env_pane) if use_env_pane else override
    yield Ribo.zoom_comp(insert_vim_pane(ident, strict, vim_pid).nvim)


__all__ = ('create_vim_pane',)
